/* Convert IFZ file to BMP (also PNG/JPG/GIF/TIF) file
 * No command line argument converts all *.ifz files in current directory.
 * Any command line argument is a file or filter name.
 * One command-line argument without wildcard will start the default BMP viewer after conversion.
   In case of multiple images, multiple instances of bitmap viewer are invoked.
 * Typically, *.ifz files should be executed with this program.
 * Ending “.ifz” will be replaced by “.bmp”, or whatever the exe file name is.
   In case of multiple cameras, a 0-based counter will be inserted, so
   example.ifz will produce example0.bmp and example1.bmp.
   In case of more than 10 cameras, a two-digit number is inserted.
   In case of multiple "inputs" (i.e. concatenated IFZ files),
   resulting image files except the first get a number in parentheses inserted:
   example.ifz will produce example.bmp, example(1).bmp, example(2).bmp and so on.
   Multiple cameras and multiple inputs are both processed independently.

Tested with monochrome camera file with 2448 x 2044 pixels.
And with color files of different sizes.
Color files are simply raw camera data from camera with specific mosaic RGB masks.
I have seen codes 10 (grayscale), 103 (RG-GB mosaic), and 104 (GR-BG mosaic).
More codes are seen by Jérôme Pinard <pinard.jerome@gmail.com>.
Note that resulting RGB files from this ifz2bmp.exe and some Omron output
differs sligtly due to unknown Omron “demosaicing” routine.
My routine is fairly simple and keeps contrast as high as possible.
If someone has an IFZ file which will be rejected, send that file to me
and a suitable screenshot (so I know how it should look after conversion).
henrik.haftmann@gmail.com
 120209	created
-120217	fixed the NULL behaviour when command-line arguments are given 
+170726	added color IFZ file conversion (after user input)
-200327	Removed dependency from first width and height values, renamed to width0 / height0
+210919	Updated IFZ header structure requires possible multiple BMP output
+211001	(losslessly compressed) PNG output using Gdiplus; also (lossy) JPG/GIF/TIF output
 */
#include <windows.h>
#ifdef UNICODE
#include <gdiplus.h>	// For now, no compression on Windows 98
#endif
#include <shlwapi.h>
#include <tchar.h>
#include <stdio.h>	// _snprintf in case of !UNICODE
#define T(x) TEXT(x)
#define elemof(x) (sizeof(x)/sizeof(*(x)))

#if _MSC_VER < 1400	// Visual C++ 6
__forceinline void __movsw(WORD*d, const WORD*s, DWORD count) {
 _asm	mov	edi,d
 _asm	mov	esi,s
 _asm	mov	ecx,count
 _asm	rep movsw
}
#pragma intrinsic(memcpy)
#else
#include <intrin.h>
#endif
#ifdef UNICODE
#define movs(t,s,l) __movsw((WORD*)(t),(WORD*)(s),l)
#else
#define movs memcpy
#endif

//target BMP definition
#pragma pack(2)
struct BMP24{
 BITMAPFILEHEADER bfh;
 BITMAPINFOHEADER bih;	// no color map; RGB colors follow
 BYTE*dataptr() {return reinterpret_cast<BYTE*>(this+1);}
};
struct BMP8:public BMP24{
 RGBQUAD colors[256];	// graymap appended; brighness bytes follow
 BYTE*dataptr() {return reinterpret_cast<BYTE*>(this+1);}
};
#pragma pack()

//source IFZ definition (as C++ class)
struct IFZIMG;
struct IFZHDR{
 DWORD signature;	// FFFFFFFE
 DWORD camdispatch;	// 00000001		210919: Camera presence bitmask
// const IFZIMG*first() const {return reinterpret_cast<const IFZIMG*>(this+1);}
};
struct IFZIMG{		// std::popcount(IFZHDR::camdispatch) structures follow:
 DWORD mosaic;		// 0000000A		170726: Encodings: See below
 DWORD camW,		// 00000990 = 2448	camera size
       camH;		// 000007FC = 2044
 DWORD offsetX,		// 00000000		offset of image inside camera size
       offsetY;		// 00000000
 DWORD imgW,		// 00000990		image width, rarely less than camW
       imgH;		// 000007FC		image height, may be less than camH+offsetY
 DWORD size;		// 004C59E0 = 5003744	entire struct size (including data[])
 const BYTE*dataptr() const {return reinterpret_cast<const BYTE*>(this+1);}
// const IFZIMG*next() const {return reinterpret_cast<const IFZIMG*>(reinterpret_cast<const BYTE*>(this)+size);}
 bool ConvertBW(BMP8&) const;
 bool ConvertRGB(BMP24&) const;
 bool Convert(BMP8&b) const {return mosaic==10?ConvertBW(b):ConvertRGB(b);}
 DWORD GetNeededBmpLength() const;
};
/* mosaic	0000000A (10)	8-bit grayscale image	monochrome cameras
		00000065 (101)	Rg-GB mosaic (+ 2021)	FHV7H-C032
		00000066 (102)	gR-BG mosaic (+ 2021)	FH-SC05R
		00000067 (103)	BG-gR mosaic		FZ-SQ050F
		00000068 (104)	GB-Rg mosaic		FH-SC21R
   101		   102		   103		   104
┌───┬───┐	┌───┬───┐	┌───┬───┐	┌───┬───┐
│ R │ g │	│ g │ R │	│ B │ G │	│ G │ B │
├───┼───┤	├───┼───┤	├───┼───┤	├───┼───┤
│ G │ B │	│ B │ G │	│ g │ R │	│ R │ g │
└───┴───┘	└───┴───┘	└───┴───┘	└───┴───┘
Every red or blue pixel is surrounded by four green pixels.
Green pixels have either red pixels above/below and blue pixels left/right (=G)
		or red pixels left/right and blue pixels above/below (=g) as neighbors.
R = 0, g = 1, G = 2, B = 3:
┌───┬───┐	┌───┬───┐	┌───┬───┐	┌───┬───┐
│ 0 │ 1 │	│ 1 │ 0 │	│ 3 │ 2 │	│ 2 │ 3 │
├───┼───┤	├───┼───┤	├───┼───┤	├───┼───┤
│ 2 │ 3 │	│ 3 │ 2 │	│ 1 │ 0 │	│ 0 │ 1 │
└───┴───┘	└───┴───┘	└───┴───┘	└───┴───┘
Xor:0		    1		    3		    2
*/
bool IFZIMG::ConvertBW(BMP8&b) const{
 ZeroMemory(&b,sizeof b);
 DWORD nPixel=size-sizeof*this;
 b.bfh.bfType='MB';
 b.bfh.bfSize=nPixel+sizeof b;	// file size
 b.bfh.bfOffBits=sizeof b;	// beyond color map
 b.bih.biSize=sizeof b.bih;
 b.bih.biWidth=imgW;
 b.bih.biHeight=imgH;
 b.bih.biPlanes=1;
 b.bih.biBitCount=8;
 for (int i=0; i<256; i++) {
  b.colors[i].rgbBlue=b.colors[i].rgbGreen=b.colors[i].rgbRed=(BYTE)i;	// fill with grayscale
 }
 const BYTE*s=dataptr();		// start of IFZ bits
 BYTE*d=b.dataptr()+nPixel;		// beyond end of BMP bits
 for (DWORD y=0; y<imgH; y++) {
  d-=imgW;
  CopyMemory(d,s,imgW);		// copy scan line
  s+=imgW;
 }
 return true;
}

bool IFZIMG::ConvertRGB(BMP24&b) const{
 ZeroMemory(&b,sizeof b);
 DWORD nPixel=size-sizeof*this;
 b.bfh.bfType='MB';
 b.bfh.bfSize=3*nPixel+sizeof b;	// file size
 b.bfh.bfOffBits=sizeof b;
 b.bih.biSize=sizeof b.bih;
 b.bih.biWidth=imgW;
 b.bih.biHeight=imgH;
 b.bih.biPlanes=1;
 b.bih.biBitCount=24;
 const BYTE*s=dataptr();		// start of IFZ bits
 BYTE *d=b.dataptr()+3*nPixel;		// beyond end of BMP bits
 int W=(int)imgW;			// W and H must be even
 int H=(int)imgH;
 int M=mosaic-101;			// zero-based mosaic enum
 BYTE B=2;	// border bit mask [l,t,r,b], start with “top”
// Putting borders into a bit mask should speed up later processing.
 for (int y=0;;y++) {
  d-=3*W;				// address BMP data bottom-up
  if (y==H-1) B|=8;			// at last line, set “bottom” mark
  B|=1;					// set “left” border mark
  for (int x=0;;x++) {			// So-called “demosaicing”
   if (x==W-1) B|=4;			// at last column, set “right” mark
   BYTE c=(y&1)<<1|x&1;			// color code for Rg-GB (see switch(c) below)
   if (M&1) c^=1;			// swap horizontally
   if (M&2) c^=3;			// swap according to table above
// Other codes can be easily added here, small “g” means green with blue above+below, not red.
   int b=0,g=0,r=0,bn=0,gn=0,rn=0;	// Prepare accumulators and divisors
   switch (c){		// For every pixel use up to 4 neighbours and average their values
    case 0: {				// A red pixel
     if (!(B&3)) b+=s[-1-W],++bn;	// take blue	from left-above
     if (!(B&6)) b+=s[ 1-W],++bn;	//		from right-above
     if (!(B&9)) b+=s[-1+W],++bn;	//		from left-below
     if (!(B&12))b+=s[ 1+W],++bn;	//		from right-below
     if (!(B&1)) g+=s[-1  ],++gn;	// take green	from left
     if (!(B&2)) g+=s[  -W],++gn;	//		from above
     if (!(B&4)) g+=s[ 1  ],++gn;	//		from right
     if (!(B&8)) g+=s[   W],++gn;	//		from below
     r=*s,++rn;				// red as-is
    }break;
    case 1: {				// A green pixel with red left+right and blue top+bottom (g)
     if (!(B&2)) b+=s[  -W],++bn;	// take blue	from above
     if (!(B&8)) b+=s[   W],++bn;	//		from below
     g=*s,++gn;				// green as-is
     if (!(B&1)) r+=s[-1  ],++rn;	// take red	from left
     if (!(B&4)) r+=s[ 1  ],++rn;	//		from right
    }break;
    case 2: {				// A green pixel with blue left+right and red top+bottom (G)
     if (!(B&1)) b+=s[-1  ],++bn;	// take blue	from left
     if (!(B&4)) b+=s[ 1  ],++bn;	//		from right
     g=*s,++gn;				// green as-is
     if (!(B&2)) r+=s[  -W],++rn;	// take red	from above
     if (!(B&8)) r+=s[   W],++rn;	//		from below
    }break;
    case 3: default: {			// A blue pixel
     b=*s,++bn;				// blue as-is
     if (!(B&1)) g+=s[-1  ],++gn;	// take green	from left
     if (!(B&2)) g+=s[  -W],++gn;	//		from above
     if (!(B&4)) g+=s[ 1  ],++gn;	//		from right
     if (!(B&8)) g+=s[   W],++gn;	//		from below
     if (!(B&3)) r+=s[-1-W],++rn;	// take red	from left-above
     if (!(B&6)) r+=s[ 1-W],++rn;	//		from right-above
     if (!(B&9)) r+=s[-1+W],++rn;	//		from left-below
     if (!(B&12))r+=s[ 1+W],++rn;	//		from right-below
    }break;
   }
   *d++=BYTE(b/bn);	// Only in case of pictures less than 4×4 pixels, divisors are zero.
   *d++=BYTE(g/gn);	// Then the program will crash intentionally.
   *d++=BYTE(r/rn);
   s++;
   if (B&4) break;			// “right” mark set? Exit loop!
   B&=~1;				// remove “left” mark
  }
  if (B&8) break;			// “bottom” mark set? Exit loop!
  B&=~6;				// remove “top” and “right” mark
  d-=3*W;				// advance
 }
 return true;
}

DWORD IFZIMG::GetNeededBmpLength() const{
 if (offsetX+imgW>camW) return 0;	// illegal
 if (offsetY+imgH>camH) return 0;	// illegal
 DWORD nPixel=imgW*imgH;
 if (nPixel>size-sizeof*this) return 0;	// Not enough data for pixels
 if (imgW&3) return 0;			// non-DWORD-aligned width is currently not implemented
 switch (mosaic) {
  case 10: return sizeof(BMP8)+nPixel;	// grayscale with palette
  case 101:
  case 102:
  case 103:
  case 104: if (imgH&1) return 0;	// cannot handle odd line count
  return sizeof(BMP24)+3*nPixel;	// RGB without palette
 }
 return 0;				// Unknown mosaic code
}

static unsigned popcount(unsigned v) {unsigned c; for (c=0; v; v>>=1) c+=v&1; return c;}

static TCHAR fileext[8]=T(".bmp");

// <i> is the current camera number out of <j> cameras.
// <k> is the sequence number ("input") of concatenated IFZ files.
// As number of inputs will be detected while decoding, the first file(s)
// don't get a "(0)" appended to file name part.
// Despite of function name, it will append <fileext> and suits for any picture format.
static TCHAR*makebmpname(const TCHAR*n,unsigned i,unsigned j, unsigned k) {
 TCHAR z[4],z2[8];
 size_t	l1=PathFindExtension(n)-n,
	l2=j>1?_sntprintf(z,elemof(z),T("%0*u"),j>10?2:1,i):0,
	l3=k?_sntprintf(z2,elemof(z2),T("(%u)"),k):0,
	l4=_tcslen(fileext);
 TCHAR*r=new TCHAR[l1+l2+l3+l4+1];
 movs(r,n,l1);
 movs(r+l1,z,l2);
 movs(r+l1+l2,z2,l3);
 movs(r+l1+l2+l3,fileext,l4+1);	// concatenate all 4 parts
 return r;
}

struct FileSize{
 DWORD lo,hi;
 operator size_t() const {return *reinterpret_cast<const size_t*>(this);}
 operator size_t&() {return *reinterpret_cast<size_t*>(this);}
};

void _cdecl message(UINT id,...) {
 TCHAR s[512],t[128];
 LoadString(0,id,t,elemof(t));
 va_list va;
 va_start(va,id);
 _vsntprintf(s,elemof(s),t,va);
 va_end(va);
 MessageBox(0,s,NULL,MB_OK|MB_ICONEXCLAMATION);
}

#ifdef UNICODE
// MUST be called, otherwise Image::Save() produces error 10!
static bool GetCodecClsid(const WCHAR*format, GUID*guid) {	// helper function
 UINT  num,size=0;	// number of image encoders, size of the image encoder array in bytes
 Gdiplus::GetImageEncodersSize(&num,&size);
 if (!size) return false;
 BYTE*mem=new BYTE[size];
 Gdiplus::ImageCodecInfo*ci=reinterpret_cast<Gdiplus::ImageCodecInfo*>(mem);
 Gdiplus::GetImageEncoders(num,size,ci);
 for(UINT j=0; j<num; ++j) if (!wcscmp(ci[j].MimeType,format)) {
  *guid=ci[j].Clsid;	// copy 16 bytes
  delete[]mem;
  return true;
 }//for+if
 delete[]mem;
 return false;
}
static wchar_t*mimetype;
#endif

static bool show;

static bool ConvertAll(PCTSTR ifzname) {
 bool ret=false;
 HANDLE fifz=CreateFile(ifzname,GENERIC_READ,0,NULL,OPEN_EXISTING,0,0);
 if (fifz!=INVALID_HANDLE_VALUE) {
  FileSize size;
  size.lo=GetFileSize(fifz,&size.hi);
  if (size) {
   HANDLE mifz=CreateFileMapping(fifz,NULL,PAGE_READONLY,0,0,NULL);
   if (mifz) {
    const BYTE*filestart=reinterpret_cast<const BYTE*>(MapViewOfFile(mifz,FILE_MAP_READ,0,0,0));
    if (filestart) {
     size_t offset=0;
     for(int k=0;offset<size;k++) {
      const IFZHDR*hdr=reinterpret_cast<const IFZHDR*>(filestart+offset);
      if (size-offset>=sizeof IFZHDR && hdr->signature==0xFFFFFFFE) {
       unsigned cams=popcount(hdr->camdispatch);
       if (cams) {
        offset+=8;
        for (unsigned i=0; i<cams; i++) {
         const IFZIMG*img=reinterpret_cast<const IFZIMG*>(filestart+offset);
	 if (size-offset>=sizeof IFZHDR && size-offset>=img->size) {
          size_t len=img->GetNeededBmpLength();
          if (len) {
           TCHAR*bmpname=makebmpname(ifzname,i,cams,k);
	   bool okay=false;
#ifdef UNICODE
	   if (mimetype) {
	    HGLOBAL hg=GlobalAlloc(GHND,len);
	    if (hg) {
	     BMP8*b=reinterpret_cast<BMP8*>(GlobalLock(hg));
             if (b) {
              if (img->Convert(*b)) {
	       IStream*memstream;
               HRESULT hr=CreateStreamOnHGlobal(hg,false,&memstream);
	       if (!hr) {
			// Initialize GDI+.
                Gdiplus::GdiplusStartupInput gdiplusStartupInput;
                ULONG_PTR gdiplusToken;
                Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
                Gdiplus::Image*image = new Gdiplus::Image(memstream);
			// CLSID of the PNG encoder. The encoder creates still quite large files!
                GUID codecGuid;//={0x557cf406,0x1a04,0x11d3,{0x73,0x9a,0x00,0x00,0xf8,0x1e,0xf3,0x2e}};
		if (GetCodecClsid(mimetype,&codecGuid)) {
                 Gdiplus::Status stat=image->Save(bmpname,&codecGuid,NULL);
                 if(stat==Gdiplus::Ok) okay=true;
		 else message(9,bmpname,stat);
		}else message(9,bmpname,0);
                delete image;
                Gdiplus::GdiplusShutdown(gdiplusToken);
		// delete memstream;	// seems that “delete image” deletes memstream
	       }else message(9,bmpname,hr);
	      }else message(8,offset);		// conversion error
	      GlobalUnlock(hg);
	     }else message(4,bmpname,len);	// GlobalAlloc is anonymous file mapping
	     GlobalFree(hg);
	    }else message(4,bmpname,len);	// GlobalAlloc is anonymous file mapping
	   }else
#endif
	   {//bmp, not png or jpg
            HANDLE fbmp=CreateFile(bmpname,GENERIC_READ|GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,0);
            if (fbmp!=INVALID_HANDLE_VALUE) {
             HANDLE mbmp=CreateFileMapping(fbmp,NULL,PAGE_READWRITE,0,DWORD(len),NULL);
             if (mbmp) {
              BMP8*b=reinterpret_cast<BMP8*>(MapViewOfFile(mbmp,FILE_MAP_WRITE,0,0,len));
              if (b) {
	       if (img->Convert(*b)) okay=true;
	       else message(8,offset);
               UnmapViewOfFile(b);
              }else message(4,bmpname,len);
              CloseHandle(mbmp);
             }else message(4,bmpname,len);
             CloseHandle(fbmp);
            }else message(6,bmpname,GetLastError());	// cannot create file
	   }//if(png)
           if (okay && show) {
	    HINSTANCE h=ShellExecute(0,NULL,bmpname,NULL,NULL,SW_SHOWDEFAULT);
	    if ((unsigned)h<32) message(7,bmpname,h);
	   }
           delete[] bmpname;
	  }else{ message(5,ifzname,offset); break;}	// wrong IFZIMG due to illegal values inside
	 }else{ message(5,ifzname,offset); break;}	// wrong IFZIMG due to too less data
	 offset+=img->size;
        }//for
       }else{ message(5,ifzname,offset); break;}	// wrong IFZHDR due to no cameras
      }else{ message(5,ifzname,offset);	break;}		// wrong IFZHDR due to wrong signature
     }//for
     UnmapViewOfFile(filestart);
    }else message(4,ifzname,size);
    CloseHandle(mifz);
   }else message(4,ifzname,size);
  }else message(3,ifzname);		// empty file
  CloseHandle(fifz);
 }else message(2,ifzname,GetLastError());	// cannot open file
 return ret;
}

static void ConvertWild(PCTSTR ifzname) {
 size_t lp=PathFindFileName(ifzname)-ifzname;	// length of path
 WIN32_FIND_DATA fd;
 HANDLE h=FindFirstFile(ifzname,&fd);
 if (h!=INVALID_HANDLE_VALUE) {
  do{
   size_t ln=_tcslen(fd.cFileName);
   TCHAR*n=new TCHAR[lp+ln+1];
   movs(n,ifzname,lp);
   movs(n+lp,fd.cFileName,ln);
   n[lp+ln]=0;
   ConvertAll(n);
  }while (FindNextFile(h,&fd));
  FindClose(h);
 }else message(10,ifzname,GetLastError());
}

#ifndef UNICODE
EXTERN_C _CRTIMP int _cdecl __getmainargs(int*,char***,char***,int);
#endif

void WinMainCRTStartup() {
#ifdef UNICODE
 int argc;
 PWSTR*argv=CommandLineToArgvW(GetCommandLine(),&argc);
#else
 __getmainargs(&__argc,&__argv,&_environ,FALSE);
 int argc=__argc;
 char**argv=__argv;
#endif
#ifdef UNICODE
 wchar_t*p2=wcsrchr(argv[0],'2');
// exe file name can be ifz2bmp, ifz2png, ifz2jpg, ifz2jpeg, ifz2tif, ifz2tiff, ifz2gif
// In case of other extensions, a suitable Gdiplus encoder must be installed.
 if (p2) {
  wchar_t*pp=wcschr(++p2,'.');
  if (!pp) pp=p2+_tcslen(p2);		// in case that exe file has no extension (possibly wine?)
  _sntprintf(fileext+1,elemof(fileext)-1,L"%.*s",pp-p2,p2);	// replace file extension
  if (wcscmp(fileext+1,L"bmp")) {	// for bmp, don't go over Gdiplus, save directly to file
   wchar_t*type=fileext+1;
   if (!wcscmp(type,L"jpg")) type=L"jpeg";	// keep typical DOS/Windows extension “jpg” but let find “jpeg” encoder
   if (!wcscmp(type,L"tif")) type=L"tiff";	// keep typical DOS/Windows extension “tif” but let find “tiff” encoder
// A clean and extensible way would be querying registry (HKCR) for file extension's mime type.
   size_t l=6+_tcslen(type)+1;		// 6 == strlen("image/")
   mimetype=new wchar_t[l];		// mimetype != nullptr: let compress using GDI+ encoders
   _snwprintf(mimetype,l,L"image/%s",type);	// build mime type for finding the encoder
  }
 }
#endif
 if (argc==2 && !_tcspbrk(argv[1],T("?*"))) show=true;
 if (argc<2) ConvertWild(T("*.ifz"));		// Without argument, convert all inside current directory
 else for (int i=1; i<argc; i++) ConvertWild(argv[i]);
// ConvertWild() will generate message boxes in case of errors.
 ExitProcess(0);
}
